package net.locusworks.s3sync.client;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.StringUtils;
import com.amazonaws.services.s3.model.S3Object;
import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.ApplicationLoggerFactory;

public class FileManager implements AutoCloseable {

  private ApplicationLogger logger = ApplicationLoggerFactory.getLogger(FileManager.class);

  public static final String FILE_CSV = "upload.csv";

  private Map<String, FileDetail> detailMap;
  
  private Set<String> s3Files;

  private S3Client client;

  private String bucket;

  private static FileManager instance;

  private FileManager(S3Client client) throws IOException {
    detailMap = new LinkedHashMap<String, FileDetail>();
    s3Files = new HashSet<String>();
    this.client = client;
    this.bucket = client.getBucket();
    readFile();
  }

  private void readFile() throws IOException {
    
    S3Object hashFile = client.getObject(bucket, FILE_CSV);
    if (hashFile == null) return;
    Path file = Paths.get(FILE_CSV);
    
    client.downloadFile(FILE_CSV, file);
    s3Files = client.getFileList();
    if (Files.notExists(file)) return;

    try(BufferedReader reader = Files.newBufferedReader(file)) {
      String line = null;
      while((line = reader.readLine()) != null) {
        String[] values = line.split(",");
        if (values.length != 3) {
          logger.warn("Invalid entry detected: " + line);
          continue;
        }

        FileDetail fd = new FileDetail(values[0], values[1], Boolean.valueOf(values[2]));
        detailMap.put(values[0], fd);
      }
    }
  }

  private void saveFile() throws IOException {
    try(BufferedWriter writer = Files.newBufferedWriter(Paths.get(FILE_CSV))) {
      writer.write("FILE_NAME,HASH,STATUS\n");
      for(FileDetail v : detailMap.values()) {
        writer.write(String.format("%s,%s,%s%n", v.getFile(), v.getHash(), v.isUploaded()));
      };
    }
  }

  public boolean addEntry(String file, String hash, boolean uploaded) {
    return addEntry(new FileDetail(file, hash, uploaded));
  }

  public boolean addEntry(FileDetail fd) {
    return detailMap.put(fd.getFile(), fd) != null;
  }

  public FileDetail getFileDetail(Path path, String key) throws IOException {
    boolean newFile = false;
    FileDetail fd = null;
    if (detailMap.containsKey(key)) {
      fd = detailMap.get(key);
    } else {
      newFile = true;
      fd = new FileDetail(key, DigestUtils.sha1Hex(Files.newInputStream(path)), false);
    }

    String sha1 = newFile ? fd.getHash() : DigestUtils.sha1Hex(Files.newInputStream(path));
    
    if (sha1.equals(fd.getHash()) && s3Files.contains(key)) {
      fd.setUploaded(true);
      s3Files.remove(key);
    } else {
      fd.setUploaded(false);
    }

    return fd;
  }
  
  public void removeOrphanedFiles() {
    client.removeFiles(s3Files);
  }

  @Override
  public void close() throws Exception {
    saveFile();
    client.uploadFile(Paths.get(FILE_CSV));
    Files.deleteIfExists(Paths.get(FILE_CSV));
  }

  public static FileManager newInstance(S3Client client) throws IOException {
    instance = new FileManager(client);
    return instance;
  }

  public static FileManager getInstance() throws IOException {
    if (instance == null || instance.client == null || StringUtils.isBlank(instance.bucket)) {
      throw new IOException("a call to newInstance() was not made. Please call newInstance first");
    }
    return instance;
  }

}
