package net.locusworks.s3sync.client;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.Set;
import com.amazonaws.AmazonClientException;
import com.amazonaws.AmazonServiceException;
import com.amazonaws.regions.Regions;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.amazonaws.services.s3.model.Bucket;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectSummary;
import com.amazonaws.services.s3.transfer.Download;
import com.amazonaws.services.s3.transfer.TransferManager;
import com.amazonaws.services.s3.transfer.TransferManagerBuilder;
import com.amazonaws.services.s3.transfer.Upload;
import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.ApplicationLoggerFactory;
import net.locusworks.s3sync.conf.ConfigurationManager;

public class S3Client implements AutoCloseable {

  private ApplicationLogger logger = ApplicationLoggerFactory.getLogger(S3Client.class);

  private AmazonS3 s3Client;
  private String bucket;
  private Path syncFolder;

  public S3Client(ConfigurationManager conf) {
    String region = conf.getRegion();
    this.s3Client = AmazonS3ClientBuilder.standard().withRegion(Regions.fromName(region)).build();
    Bucket bucket = s3Client.listBuckets().stream()
        .filter(b -> b.getName().equals(conf.getBucketName())).findFirst().orElse(null);

    if (bucket == null) {
      logger.error("Unable to find bucket with name: %s%nExiting", conf.getBucketName());
      System.exit(-1);
    }
    logger.info("Found Bucket: %s", bucket);
    this.bucket = conf.getBucketName();
    this.syncFolder = conf.getSyncFolder();
  }
  
  public String getBucket() {
    return this.bucket;
  }

  public void uploadFile(Path file) {
    TransferManager xferMgr = TransferManagerBuilder.standard().withS3Client(s3Client).build();
    uploadFile(xferMgr, file);
    xferMgr.shutdownNow(false);
  }

  public void uploadFile(TransferManager xferMgr, Path file) {
    boolean xferMgrNull = xferMgr == null;
    xferMgr = !xferMgrNull ? xferMgr : TransferManagerBuilder.standard().withS3Client(s3Client).build();
    FileDetail fd = null;
    try {
      String key = getPath(file);
      fd = FileManager.getInstance().getFileDetail(file, key);
      if (fd.isUploaded()) return;
      logger.info("Uploading file: %s", file);
      Upload xfer = xferMgr.upload(bucket, key, file.toFile());
      xfer.waitForCompletion();
      fd.setUploaded(true);
      FileManager.getInstance().addEntry(fd);
      logger.info("Done uploading %s", file);
    } catch (AmazonClientException | InterruptedException | IOException e) {
      if (fd != null) {
        fd.setUploaded(false);
        try {
          FileManager.getInstance().addEntry(fd);
        } catch (IOException e1) {
          logger.error("Unable to save file to file manager: " + e1.getMessage());
        }
      }
      logger.error(e.getMessage());
    } finally {
      if (xferMgrNull) {
        xferMgr.shutdownNow(false);
      }
    }
  }
  
  public void downloadFile(String key, Path file) {
    TransferManager xferMgr = TransferManagerBuilder.standard().withS3Client(s3Client).build();
    downloadFile(xferMgr, key, file);
    xferMgr.shutdownNow(false);
  }

  public void downloadFile(TransferManager xferMgr, String key, Path file) {
    boolean xferMgrNull = xferMgr == null;
    xferMgr = !xferMgrNull ? xferMgr : TransferManagerBuilder.standard().withS3Client(s3Client).build();
    try {
      logger.info("Downloading file: %s", file);
      Download xfer = xferMgr.download(bucket, key, file.toFile());
      xfer.waitForCompletion();
      logger.info("Done downloading %s", file);
    } catch (AmazonClientException | InterruptedException e) {
      logger.error(e.getMessage());
    } finally {
      if (xferMgrNull) {
        xferMgr.shutdownNow(false);
      }
    }
  }

  public void syncFolder() throws IOException {
    TransferManager xferMgr = TransferManagerBuilder.standard().withS3Client(s3Client).build();
    try (FileManager manager = FileManager.getInstance()) {
      Files.walk(syncFolder)
      .filter(f -> Files.isRegularFile(f))
      .forEach(f -> uploadFile(xferMgr, syncFolder.resolve(f)));
    } catch (Exception e) {
      logger.error("Unable to load file Manager: " + e.getMessage());
    } 

    xferMgr.shutdownNow(false);
  }
  
  public S3Object getObject(String bucketName, String key) {
    try {
      return s3Client.getObject(bucketName, key);
    } catch (AmazonServiceException ex) {}
    return null;
  }
  
  public Set<String> getFileList() {
    Set<String> fileList = new HashSet<String>();
    for(S3ObjectSummary os: s3Client.listObjectsV2(getBucket()).getObjectSummaries()) {
      fileList.add(os.getKey());
    }
    return fileList;
  }
  
  public void removeFiles(Set<String> s3Files) {
    for(String key : s3Files) {
      try {
        logger.info("Removing file: " + key);
        s3Client.deleteObject(getBucket(), key);
      } catch (AmazonServiceException ex) {
        logger.warn(String.format("Unable to delete %s: %s", key, ex.getMessage()), ex);
      }
    }
  }
 
  private String getPath(Path file) {
    if (file.getParent() == null) return file.getFileName().toString();
    Path relative = syncFolder.relativize(file);
    return relative.toString().replace("\\", "/");
  }
  
  @Override
  public void close() throws Exception {
    if (s3Client != null) {
      s3Client.shutdown();
    }
  }

}
